//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import Alphabet;
import Distribution;
import State;
import MarkovModel;
import Common;
import Estimator;
import JointCNVDistribution;
import tango.math.Math;

//
// Create the structure of the model but dont attach meaningful distributions yet
//
MarkovModel initializeModel()
{
	// Create the model
	DiscreteAlphabet da = new DiscreteAlphabet;
	da.createSymbol("het");
	da.createSymbol("depth");
	da.createSymbol("avghit");
	MarkovModel model = new MarkovModel(da);

	// Create the states with default emission distributions
	State s_zero = new State(SNAME_ZERO, new EmptyDistribution());
	State s_norm = new State(SNAME_NORMAL, new EmptyDistribution());
	State s_cnHalf = new State(SNAME_HALF, new EmptyDistribution());
	State s_cnTwo = new State(SNAME_C2, new EmptyDistribution());
	State s_cnThree = new State(SNAME_C3, new EmptyDistribution());
	State s_repeat = new State(SNAME_REPEAT, new EmptyDistribution());

	// Add the states to the model
	model.addState(s_zero);
	model.addState(s_norm);
	model.addState(s_cnHalf);
	model.addState(s_cnTwo);
	model.addState(s_cnThree);
	model.addState(s_repeat);

	// Setup the transitions
	State terminalState = model.getTerminalState();

	// Terminal State transitions
	model.createTransition(terminalState, s_repeat, 1.0);	
	model.createTransition(terminalState, s_zero, 5.0);
	model.createTransition(terminalState, s_norm, 5.0);
	model.createTransition(terminalState, s_cnHalf, 1.0);
	model.createTransition(terminalState, s_cnTwo, 1.0);
	model.createTransition(terminalState, s_cnThree, 1.0);
	model.normalizeTransitions(terminalState);

	// repeat state transitions
	model.createTransition(s_repeat, s_repeat, 1000.0);	
	model.createTransition(s_repeat, s_zero, 1.0);
	model.createTransition(s_repeat, s_norm, 1.0);
	model.createTransition(s_repeat, s_cnHalf, 1.0);
	model.createTransition(s_repeat, s_cnTwo, 1.0);
	model.createTransition(s_repeat, s_cnThree, 1.0);
	model.createTransition(s_repeat, terminalState, 1.0);
	model.normalizeTransitions(s_repeat);
	

	// zero state transitions
	model.createTransition(s_zero, s_repeat, 1.0);	
	model.createTransition(s_zero, s_zero, 1000.0);
	model.createTransition(s_zero, s_norm, 1.0);
	model.createTransition(s_zero, s_cnHalf, 1.0);
	model.createTransition(s_zero, s_cnTwo, 1.0);
	model.createTransition(s_zero, s_cnThree, 1.0);
	model.createTransition(s_zero, terminalState, 1.0);
	model.normalizeTransitions(s_zero);

	// normal state transitions
	model.createTransition(s_norm, s_repeat, 1.0);	
	model.createTransition(s_norm, s_zero, 1.0);
	model.createTransition(s_norm, s_norm, 100000.0);
	model.createTransition(s_norm, s_cnHalf, 1.0);
	model.createTransition(s_norm, s_cnTwo, 1.0);
	model.createTransition(s_norm, s_cnThree, 1.0);
	model.createTransition(s_norm, terminalState, 1.0);
	model.normalizeTransitions(s_norm);

	// cnHalf
	model.createTransition(s_cnHalf, s_repeat, 1.0);	
	model.createTransition(s_cnHalf, s_zero, 1.0);
	model.createTransition(s_cnHalf, s_norm, 1.0);
	model.createTransition(s_cnHalf, s_cnHalf, 1000.0);
	model.createTransition(s_cnHalf, s_cnTwo, 1.0);
	model.createTransition(s_cnHalf, s_cnThree, 1.0);
	model.createTransition(s_cnHalf, terminalState, 1.0);
	model.normalizeTransitions(s_cnHalf);

	// cnTwo
	model.createTransition(s_cnTwo, s_repeat, 1.0);	
	model.createTransition(s_cnTwo, s_zero, 1.0);
	model.createTransition(s_cnTwo, s_norm, 1.0);
	model.createTransition(s_cnTwo, s_cnHalf, 1.0);
	model.createTransition(s_cnTwo, s_cnTwo, 1000.0);
	model.createTransition(s_cnTwo, s_cnThree, 1.0);
	model.createTransition(s_cnTwo, terminalState, 1.0);
	model.normalizeTransitions(s_cnTwo);

	// cnThree
	model.createTransition(s_cnThree, s_repeat, 1.0);	
	model.createTransition(s_cnThree, s_zero, 1.0);
	model.createTransition(s_cnThree, s_norm, 1.0);
	model.createTransition(s_cnThree, s_cnHalf, 1.0);
	model.createTransition(s_cnThree, s_cnTwo, 1.0);
	model.createTransition(s_cnThree, s_cnThree, 1000.0);
	model.createTransition(s_cnThree, terminalState, 1.0);
	model.normalizeTransitions(s_cnThree);
	
	return model;
}

// update the emission distributions for the model according to the parameters
void updateModelDistributions(MarkovModel model, ModelParameters parameters)
{
	// Create the het emission distributsion from the parameters
	IDistribution zeroHet = new PoissonDistribution(parameters.stateParams[SNAME_ZERO].het_mean);
	IDistribution normalHet = new PoissonDistribution(parameters.stateParams[SNAME_NORMAL].het_mean);
	IDistribution cnHalfHet = new PoissonDistribution(parameters.stateParams[SNAME_HALF].het_mean);
	IDistribution cnTwoHet = createExponential(parameters.stateParams[SNAME_C2].het_mean, parameters.exp_step);
	IDistribution cnThreeHet = createExponential(parameters.stateParams[SNAME_C3].het_mean, parameters.exp_step);
	IDistribution repeatHet = new PoissonDistribution(parameters.stateParams[SNAME_REPEAT].het_mean);

	// Create the depth emission distribution
	IDistribution zeroDepth = new PoissonDistribution(parameters.stateParams[SNAME_ZERO].depth_mean);
	IDistribution normalDepth = createMixedGaussian(parameters.stateParams[SNAME_NORMAL].depth_mean, parameters.normal_step, 0.01, 0.97, 0.01, 0.01);
	IDistribution cnHalfDepth = createMixedGaussian(parameters.stateParams[SNAME_NORMAL].depth_mean, parameters.normal_step, 0.97, 0.01, 0.01, 0.01);
	IDistribution cnTwoDepth = createMixedGaussian(parameters.stateParams[SNAME_NORMAL].depth_mean, parameters.normal_step, 0.01, 0.01, 0.97, 0.01);
	IDistribution cnThreeDepth = createMixedGaussian(parameters.stateParams[SNAME_NORMAL].depth_mean, parameters.normal_step, 0.01, 0.01, 0.01, 0.97);
	IDistribution repeatDepth = createExponential(parameters.stateParams[SNAME_NORMAL].depth_mean, parameters.exp_step);
	
	// Create the hit emission distribution
	IDistribution zeroHit = new PoissonDistribution(parameters.stateParams[SNAME_ZERO].avghit_mean);
	IDistribution normalHit = createExponential(parameters.stateParams[SNAME_NORMAL].avghit_mean, parameters.exp_step);
	IDistribution cnHalfHit = createExponential(parameters.stateParams[SNAME_HALF].avghit_mean, parameters.exp_step);
	IDistribution cnTwoHit = createExponential(parameters.stateParams[SNAME_C2].avghit_mean, parameters.exp_step);
	IDistribution cnThreeHit = createExponential(parameters.stateParams[SNAME_C3].avghit_mean, parameters.exp_step);
	IDistribution repeatHit = new QuantizedDistribution(parameters.normal_step, new GaussianDistribution(parameters.stateParams[SNAME_REPEAT].avghit_mean, 10));

	// Create the joint distributions
	IDistribution zeroJoint = new JointCNVDistribution(zeroHet, zeroDepth, zeroHit);
	IDistribution normalJoint =  new JointCNVDistribution(normalHet, normalDepth, normalHit);
	IDistribution cnHalfJoint =  new JointCNVDistribution(cnHalfHet, cnHalfDepth, cnHalfHit);
	IDistribution cnTwoJoint =  new JointCNVDistribution(cnTwoHet, cnTwoDepth, cnTwoHit);
	IDistribution cnThreeJoint =  new JointCNVDistribution(cnThreeHet, cnThreeDepth, cnThreeHit);
	IDistribution repeatJoint =  new JointCNVDistribution(repeatHet, repeatDepth, repeatHit);

	// Attach the distributions to the states
	model.getState(SNAME_ZERO).setEmissionDistribution(zeroJoint);
	model.getState(SNAME_NORMAL).setEmissionDistribution(normalJoint);
	model.getState(SNAME_HALF).setEmissionDistribution(cnHalfJoint);
	model.getState(SNAME_C2).setEmissionDistribution(cnTwoJoint);
	model.getState(SNAME_C3).setEmissionDistribution(cnThreeJoint);
	model.getState(SNAME_REPEAT).setEmissionDistribution(repeatJoint);
}

//
// Create a 4-way mixed gaussian depth dist
//
QuantizedDistribution createMixedGaussian(real norm_mean, real step, real halfW, real oneW, real twoW, real threeW)
{
	// TODO: More accurate model of standard deviation?
	MixedDistribution md = new MixedDistribution();
	md.addDistribution(halfW, new GaussianDistribution(0.5*norm_mean, sqrt(0.5*norm_mean)));
	md.addDistribution(oneW, new GaussianDistribution(norm_mean, sqrt(norm_mean)));
	md.addDistribution(twoW, new GaussianDistribution(2*norm_mean, sqrt(3*norm_mean)));
	md.addDistribution(threeW, new GaussianDistribution(3*norm_mean, sqrt(4*norm_mean)));
	return new QuantizedDistribution(step, md);
}

//
// Create a quantizied exponential
//
QuantizedDistribution createExponential(real exp_mean, real exp_step)
{
	return new QuantizedDistribution(exp_step, new ExponentialDistribution(1.0 / exp_mean));
}
