//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import cnD;
import CNVModel;
import Common;
import Alphabet;
import Distribution;
import JointCNVDistribution;
import tango.math.Math;
import tango.io.Stdout;

alias real[][char[]] ColumnArrayLUT;


class ModelParameters
{

	this(MarkovModel m)
	{
		// Initialize all the states
		StateParameters sp;
		foreach(s; m.getStateList())
		{
			stateParams[s.toString] = sp;
		}
	}


	void print()
	{
		foreach(l,s; stateParams)
		{
			Stdout.format("{}:\t\t{}\n", l, s.toString());
		}
		Stdout.format("Step: [{0:F2} {1:F2}]\n", normal_step, exp_step);
	}
	

	// Per-state parameters
	StateParameters[char[]] stateParams;

	// Distribution parameters
	real normal_step = 0.5;
	real exp_step = 1.0;	
}


ModelParameters generateInitialParameters(MarkovModel model, Symbol[] symbols, int windowSize)
{
	ModelParameters modelParams = new ModelParameters(model);

	// Calculate the initial depth parameters
	Statistics depthStats = calculateInitialDepth(symbols, 0.30);

	// Set initial depth params
	State[] states = model.getStateList();
	foreach(s; states)
	{
		modelParams.stateParams[s.toString()].depth_mean = depthStats.mean;
		modelParams.stateParams[s.toString()].depth_sd = depthStats.sd;
	}

	modelParams.stateParams[SNAME_ZERO].depth_mean = 0.3;
	modelParams.stateParams[SNAME_ZERO].depth_sd = 0.3;



	// Estimate the background het rate from the symbol list
	real backgroundHet = calculateInitialBackgroundHet(symbols, windowSize);
	modelParams.stateParams[SNAME_NORMAL].het_mean = backgroundHet;
	modelParams.stateParams[SNAME_ZERO].het_mean = backgroundHet;
	modelParams.stateParams[SNAME_REPEAT].het_mean = backgroundHet;
	modelParams.stateParams[SNAME_HALF].het_mean = backgroundHet;

	// For the cn gain states, we set the prior as 1 difference per copy per 1bp
	real gain_het_mean = cast(real)windowSize / 1000.0f;
	modelParams.stateParams[SNAME_C2].het_mean = (2.0 * gain_het_mean) + backgroundHet;
	modelParams.stateParams[SNAME_C3].het_mean = (3.0 * gain_het_mean) + backgroundHet;

	// Average hit priors
	real unique_hit_val = 6;
	real repeat_hit_mean = 50;
	modelParams.stateParams[SNAME_NORMAL].avghit_mean = unique_hit_val;
	modelParams.stateParams[SNAME_ZERO].avghit_mean = unique_hit_val;
	modelParams.stateParams[SNAME_REPEAT].avghit_mean = repeat_hit_mean;
	modelParams.stateParams[SNAME_HALF].avghit_mean = unique_hit_val;
	modelParams.stateParams[SNAME_C2].avghit_mean = unique_hit_val;
	modelParams.stateParams[SNAME_C3].avghit_mean = unique_hit_val;

	return modelParams;
}

// Calculate the initial estimate for the mean depth of the 1-copy state
// For the initial mean depth estimate the trimmed mean is used
Statistics calculateInitialDepth(Symbol[] symbols, real trim_percent)
{
	// Extract the depth column
	int length = symbols.length; 

	real[] depthData;
	depthData.length = length;
	
	int idx = 0;	
	foreach(s; symbols)
	{
		JointSymbol js = cast(JointSymbol)s;
		depthData[idx++] = js.depthSymbol.value;		
	}
	
	depthData.sort;

	// Calculate how much to trim
	int trimmed_length = rndint((cast(real)length * (1.0f- trim_percent)));
	int trim_start = rndint(cast(real)length * (trim_percent/2));	
	
	// Perform the trim
	real[] trimmed;
	trimmed.length = trimmed_length;
	int count = 0;
	for(int i = 0; i < depthData.length; ++i)
	{
		if(i > trim_start && count < trimmed_length)
		{
			trimmed[count++] = depthData[i];
		}
	}

	// Compute the mean over the trimmed data as the estimate
	Statistics stats = computeStats(trimmed);
	
	// Set the initial model parameters
	return stats;
}


// A collection of functions to estimate model parameters
// Calculate the initial het rate paramter
real calculateInitialBackgroundHet(Symbol[] symbols, int windowSize)
{
	// Extract the depth column
	int length = symbols.length; 

	real[] hetData;
	hetData.length = length;
	
	// Calculate the het rate parameters over the discrete
	int idx = 0;	
	foreach(s; symbols)
	{
		JointSymbol js = cast(JointSymbol)s;
		hetData[idx++] = js.hetSymbol.value;		
	}

	Statistics hetStats = computeStats(hetData);

	// Bin the data, essentially creating the empirical poisson distribution
	BinData binned = new BinData(hetData, 1.0);
	
	// The assumption for the background het state is that the vast majority of the
	// 0-het emissions are from the background
	real background_het = -log(binned.getProportion(0.0));
	return background_het;
}

//
// Create model parameters from a previous classification
//
ModelParameters createModelParametersFromClassifications(MarkovModel model, SymbolArray[State] stateSymbols)
{
	ModelParameters parameters = new ModelParameters(model);
	
	//Estimate depth
	State[] states = model.getStateList();
	foreach(s; states)
	{
		Statistics depthStats = getStatsForStateColumn(s.toString(), DEPTH_COL_NAME, model, stateSymbols);
		parameters.stateParams[s.toString()].depth_mean = depthStats.mean;
		parameters.stateParams[s.toString()].depth_sd = depthStats.sd;
	}

	// Estimate the background het rate
	real backgroundHet = getStatsForStateColumn(SNAME_NORMAL, HET_COL_NAME, model, stateSymbols).mean;
	parameters.stateParams[SNAME_NORMAL].het_mean= backgroundHet;
	parameters.stateParams[SNAME_ZERO].het_mean = backgroundHet;
	parameters.stateParams[SNAME_HALF].het_mean = backgroundHet;
	parameters.stateParams[SNAME_REPEAT].het_mean = backgroundHet;

	real c2_het = getStatsForStateColumn(SNAME_C2, HET_COL_NAME, model, stateSymbols).mean;
	real c3_het = getStatsForStateColumn(SNAME_C3, HET_COL_NAME, model, stateSymbols).mean;


	// Calculate the base het parameter by using all the copy-gain states
	real het_param = ((c2_het/2.0) + (c3_het/3.0))/3.0;

	parameters.stateParams[SNAME_C2].het_mean = 2.0 * het_param;
	parameters.stateParams[SNAME_C3].het_mean = 3.0 * het_param;

	// Calculate the avghit stats
	foreach(s; states)
	{
		parameters.stateParams[s.toString()].avghit_mean = getStatsForStateColumn(s.toString(), AVGHIT_COL_NAME, model, stateSymbols).mean;
	}
	
	return parameters;
}

//
// Estimate the baseline (normal state) depth from the classifications 
//
Statistics getStatsForStateColumn(char[] stateName, char[] colName, MarkovModel model, SymbolArray[State] stateSymbols)
{
	// The baseline depth is just the mean depth for the "normal" state

	// Extract the symbols
	real[] depthArray;
	real[] hetArray;
	real[] hitArray;
	ColumnArrayLUT caLUT = extractValueArraysForState(stateName, model, stateSymbols, depthArray, hetArray, hitArray);

	return computeStats(caLUT[colName]);
}

//
// Estimate the background het rate from the classifications
//
real estimateHetForState(char[] stateName, MarkovModel model, SymbolArray[State] stateSymbols)
{
	// The background het rate is just the mean number of hets for the "normal" state

	// Extract the symbols
	real[] depthArray;
	real[] hetArray;
	real[] hitArray;
	extractValueArraysForState(stateName, model, stateSymbols, depthArray, hetArray, hitArray);

	Statistics het = computeStats(hetArray);
	return het.mean;
}



//
// Compute statistics from an array of values
//
Statistics computeStats(in real[] values)
{
	Statistics ret;
	
	values.sort;

	if(values.length == 0)
	{
		ret.mean = 0;
		ret.median = 0;
		ret.sd = 0;
		return ret;
	}

	real half_idx = cast(real)values.length / 2;
	if(values.length % 2 == 0)
	{
		// even number of elems
		ret.median = (values[rndint(half_idx)] + values[rndint(half_idx) - 1]) / 2;
	}
	else
	{
		ret.median = values[rndint(floor(half_idx))];
	}

	// Calculate the mean
	real sum = 0.0f;
	foreach(v; values)
	{
		sum += v;
	}
	ret.mean = sum / values.length;

	// Calculate the std dev
	sum = 0;
	foreach(v; values)
	{
		sum += pow((v - ret.mean), 2.0);
	}

	ret.sd = sqrt(sum / values.length);
	return ret;
}

//
// Extract the value arrays from a StateSymbol classification for a state
//
ColumnArrayLUT extractValueArraysForState(char[] stateName, MarkovModel model, SymbolArray[State] stateSymbols, 
						out real[] depthArray, out real[] hetArray, out real[] avgHitArray)
{
	State s = model.getState(stateName);
	SymbolArray sa = stateSymbols[s];
	splitSymbolArray(sa, depthArray, hetArray, avgHitArray);

	ColumnArrayLUT stateColumns;
	stateColumns[DEPTH_COL_NAME] = depthArray;
	stateColumns[HET_COL_NAME] = hetArray;
	stateColumns[AVGHIT_COL_NAME] = avgHitArray;

	return stateColumns;
}

//
// Split an array of symbols into n-arrays of its constituent parts
//
void splitSymbolArray(SymbolArray symbolArray, out real[] depthArray, out real[] hetArray, out real[] avgHitArray)
{
	depthArray.length = symbolArray.length();
	hetArray.length = symbolArray.length();
	avgHitArray.length = symbolArray.length();

	int idx = 0;
	foreach(s; symbolArray)
	{
		// Cast the symbol to the correct type
		JointSymbol js = cast(JointSymbol)s;
		depthArray[idx] = js.depthSymbol.value;
		hetArray[idx] = js.hetSymbol.value;
		avgHitArray[idx] = js.hitSymbol.value;
		++idx;
	}	
}


//
// Simple function to calculate the probablity of getting k occurences from a poisson dist with mean = m
//
real calculatePoissonProb(int k, real m)
{
	return exp(-m) * pow(m, k) / factorial(k);
}
